@@ -33,8 +33,11 @@ from django.core.exceptions import PermissionDenied
 from django.db import transaction, models
 from django.contrib.admin.util import unquote
 from django.contrib.admin import helpers
-from django.utils.translation import ugettext as _
-from django.utils.encoding import force_unicode
+from django.utils.translation import ugettext_lazy as _
+try:
+    from django.utils.encoding import force_text
+except: # 1.4 compat
+    from django.utils.encoding import force_unicode as force_text
 from django.utils.safestring import mark_safe
 from django.contrib.admin.views.decorators import staff_member_required
 from django.http import Http404, HttpResponseRedirect
@@ -73,7 +76,11 @@ class MassAdmin(admin.ModelAdmin):
             self.admin_obj = admin_site._registry[model]
         except KeyError:
             raise Exception('Model not registered with the admin site.')
-        for (varname, var) in self.admin_obj.__class__.__dict__.iteritems():
+        try:
+            admin_obj_items = self.admin_obj.__class__.__dict__.iteritems()
+        except AttributeError:  # Python 2.x compat
+            admin_obj_items = self.admin_obj.__class__.__dict__.items()
+        for (varname, var) in admin_obj_items:
             if not (varname.startswith('_') or callable(var)):
                 self.__dict__[varname] = var
         super(MassAdmin, self).__init__(model, admin_site)
@@ -84,7 +91,7 @@ class MassAdmin(admin.ModelAdmin):
         """
         opts = obj._meta
 
-        msg = _('Selected %(name)s were changed successfully.') % {'name': force_unicode(opts.verbose_name_plural), 'obj': force_unicode(obj)}
+        msg = _('Selected %(name)s were changed successfully.') % {'name': force_text(opts.verbose_name_plural), 'obj': force_text(obj)}
 
         self.message_user(request, msg)
         return HttpResponseRedirect("../../%s/" % self.model._meta.module_name)
@@ -121,13 +128,14 @@ class MassAdmin(admin.ModelAdmin):
         general_error = None
 
         # Allow model to hide some fields for mass admin
-        self.exclude = getattr(self.admin_obj, "massadmin_exclude", None)
+        exclude_fields = getattr(self.admin_obj, "massadmin_exclude", ())
+        queryset = getattr(self.admin_obj, "massadmin_queryset", self.queryset)(request)
                         		
         object_ids = comma_separated_object_ids.split(',')
         object_id = object_ids[0]
 
         try:
-            obj = self.queryset(request).get(pk=unquote(object_id))
+            obj = queryset.get(pk=unquote(object_id))
         except model.DoesNotExist:
             obj = None
 
@@ -135,7 +143,7 @@ class MassAdmin(admin.ModelAdmin):
             raise PermissionDenied
 
         if obj is None:
-            raise Http404(_('%(name)s object with primary key %(key)r does not exist.') % {'name': force_unicode(opts.verbose_name), 'key': escape(object_id)})
+            raise Http404(_('%(name)s object with primary key %(key)r does not exist.') % {'name': force_text(opts.verbose_name), 'key': escape(object_id)})
 
         if request.method == 'POST' and request.POST.has_key("_saveasnew"):
             return self.add_view(request, form_url='../add/')
@@ -148,7 +156,7 @@ class MassAdmin(admin.ModelAdmin):
                 try:
                     objects_count = 0
                     changed_count = 0
-                    objects = self.queryset(request).filter(pk__in = object_ids)
+                    objects = queryset.filter(pk__in = object_ids)
                     for obj in objects:
                         objects_count += 1
                         form = ModelForm(request.POST, request.FILES, instance=obj)
@@ -208,7 +216,7 @@ class MassAdmin(admin.ModelAdmin):
             formset = FormSet(instance=obj, prefix=prefix)
             formsets.append(formset)
             
-        adminForm = helpers.AdminForm(form, self.get_fieldsets(request, obj), self.prepopulated_fields, self.get_readonly_fields(request, obj))
+        adminForm = helpers.AdminForm(form, self.get_fieldsets(request, obj), self.prepopulated_fields, self.get_readonly_fields(request, obj), model_admin=self.admin_obj)
         media = self.media + adminForm.media
         
         # We don't want the user trying to mass change unique fields!
@@ -229,11 +237,12 @@ class MassAdmin(admin.ModelAdmin):
         #    media = media + inline_admin_formset.media
             
         context = {
-            'title': _('Change %s') % force_unicode(opts.verbose_name),
+            'title': _('Change %s') % force_text(opts.verbose_name),
             'adminform': adminForm,
             'object_id': object_id,
             'original': obj,
             'unique_fields': unique_fields,
+            'exclude_fields': exclude_fields,
             'is_popup': request.REQUEST.has_key('_popup'),
             'media': mark_safe(media),
             #'inline_admin_formsets': inline_admin_formsets,
